@@ -26,36 +26,32 @@ import java.util.Collections;
 import java.util.Dictionary;
 import java.util.Enumeration;
 import java.util.Hashtable;
-import java.util.LinkedHashMap;
 import java.util.List;
-import java.util.Map;
-import java.util.concurrent.Callable;
-import java.util.concurrent.ExecutorService;
-import java.util.concurrent.Executors;
-import java.util.concurrent.ThreadFactory;
-import java.util.concurrent.TimeUnit;
 
-import javax.servlet.ServletContext;
 import javax.servlet.SessionCookieConfig;
 import javax.servlet.SessionTrackingMode;
 
-import org.apache.felix.http.base.internal.DispatcherServlet;
-import org.apache.felix.http.base.internal.EventDispatcher;
 import org.apache.felix.http.base.internal.HttpServiceController;
 import org.apache.felix.http.base.internal.logger.SystemLogger;
+import org.apache.felix.http.jetty.internal.webapp.WebAppBundleTracker;
+import org.eclipse.jetty.alpn.server.ALPNServerConnectionFactory;
 import org.eclipse.jetty.http.HttpVersion;
+import org.eclipse.jetty.http2.server.HTTP2ServerConnectionFactory;
+import org.eclipse.jetty.io.ConnectionStatistics;
 import org.eclipse.jetty.security.HashLoginService;
+import org.eclipse.jetty.security.UserStore;
 import org.eclipse.jetty.server.Connector;
-import org.eclipse.jetty.server.ConnectorStatistics;
 import org.eclipse.jetty.server.HttpConfiguration;
 import org.eclipse.jetty.server.HttpConnectionFactory;
 import org.eclipse.jetty.server.SecureRequestCustomizer;
 import org.eclipse.jetty.server.Server;
 import org.eclipse.jetty.server.ServerConnector;
-import org.eclipse.jetty.server.SessionManager;
 import org.eclipse.jetty.server.SslConnectionFactory;
 import org.eclipse.jetty.server.handler.ContextHandlerCollection;
 import org.eclipse.jetty.server.handler.StatisticsHandler;
+import org.eclipse.jetty.server.handler.gzip.GzipHandler;
+import org.eclipse.jetty.server.session.HouseKeeper;
+import org.eclipse.jetty.server.session.SessionHandler;
 import org.eclipse.jetty.servlet.ServletContextHandler;
 import org.eclipse.jetty.servlet.ServletHolder;
 import org.eclipse.jetty.util.component.AbstractLifeCycle;
@@ -65,70 +61,52 @@ import org.eclipse.jetty.util.thread.QueuedThreadPool;
 import org.eclipse.jetty.util.thread.ThreadPool;
 import org.osgi.framework.Bundle;
 import org.osgi.framework.BundleContext;
-import org.osgi.framework.BundleEvent;
 import org.osgi.framework.Constants;
-import org.osgi.framework.ServiceReference;
+import org.osgi.framework.InvalidSyntaxException;
+import org.osgi.framework.ServiceFactory;
 import org.osgi.framework.ServiceRegistration;
-import org.osgi.service.event.Event;
-import org.osgi.service.event.EventAdmin;
 import org.osgi.service.http.runtime.HttpServiceRuntimeConstants;
-import org.osgi.util.tracker.BundleTracker;
-import org.osgi.util.tracker.BundleTrackerCustomizer;
-import org.osgi.util.tracker.ServiceTracker;
-import org.osgi.util.tracker.ServiceTrackerCustomizer;
 
-public final class JettyService extends AbstractLifeCycle.AbstractLifeCycleListener implements BundleTrackerCustomizer, ServiceTrackerCustomizer
+public final class JettyService extends AbstractLifeCycle.AbstractLifeCycleListener
 {
     /** PID for configuration of the HTTP service. */
     public static final String PID = "org.apache.felix.http";
 
-    private static final String HEADER_WEB_CONTEXT_PATH = "Web-ContextPath";
-    private static final String HEADER_ACTIVATION_POLICY = "Bundle-ActivationPolicy";
-    private static final String WEB_SYMBOLIC_NAME = "osgi.web.symbolicname";
-    private static final String WEB_VERSION = "osgi.web.version";
-    private static final String WEB_CONTEXT_PATH = "osgi.web.contextpath";
-    private static final String OSGI_BUNDLE_CONTEXT = "osgi-bundlecontext";
-
     private final JettyConfig config;
     private final BundleContext context;
-    private final DispatcherServlet dispatcher;
     private final HttpServiceController controller;
-    private final Map<String, Deployment> deployments;
-    private final ExecutorService executor;
 
-    private ServiceRegistration<?> configServiceReg;
-    private Server server;
-    private ContextHandlerCollection parent;
-    private EventDispatcher eventDispatcher;
-    private MBeanServerTracker mbeanServerTracker;
-    private BundleTracker bundleTracker;
-    private ServiceTracker eventAdmintTracker;
-    private ServiceTracker connectorTracker;
-    private ServiceTracker loadBalancerCustomizerTracker;
-    private CustomizerWrapper customizerWrapper;
-    private EventAdmin eventAdmin;
+    private volatile ServiceRegistration<?> configServiceReg;
+    private volatile Server server;
+    private volatile ContextHandlerCollection parent;
+    private volatile MBeanServerTracker mbeanServerTracker;
+    private volatile ConnectorFactoryTracker connectorTracker;
+    private volatile RequestLogTracker requestLogTracker;
+    private volatile LogServiceRequestLog osgiRequestLog;
+    private volatile FileRequestLog fileRequestLog;
+    private volatile LoadBalancerCustomizerFactoryTracker loadBalancerCustomizerTracker;
+    private volatile CustomizerWrapper customizerWrapper;
+    private boolean registerManagedService = true;
+    private final String jettyVersion;
+    volatile WebAppBundleTracker webAppTracker;
 
     public JettyService(final BundleContext context,
-            final DispatcherServlet dispatcher,
-            final EventDispatcher eventDispatcher,
             final HttpServiceController controller)
     {
+        this.jettyVersion = fixJettyVersion(context);
+
         this.context = context;
         this.config = new JettyConfig(this.context);
-        this.dispatcher = dispatcher;
-        this.eventDispatcher = eventDispatcher;
         this.controller = controller;
-        this.deployments = new LinkedHashMap<String, Deployment>();
-        this.executor = Executors.newSingleThreadExecutor(new ThreadFactory()
-        {
-            @Override
-            public Thread newThread(Runnable runnable)
-            {
-                Thread t = new Thread(runnable);
-                t.setName("Jetty HTTP Service");
-                return t;
-            }
-        });
+    }
+
+    public JettyService(final BundleContext context,
+            final HttpServiceController controller,
+            final Dictionary<String,?> props)
+    {
+        this(context, controller);
+   	    this.config.update(props);
+   	    this.registerManagedService = false;
     }
 
     public void start() throws Exception
@@ -136,15 +114,37 @@ public final class JettyService extends AbstractLifeCycle.AbstractLifeCycleListe
         // FELIX-4422: start Jetty synchronously...
         startJetty();
 
-        Dictionary<String, Object> props = new Hashtable<String, Object>();
-        props.put(Constants.SERVICE_PID, PID);
-        this.configServiceReg = this.context.registerService("org.osgi.service.cm.ManagedService", new JettyManagedService(this), props);
+        if (this.registerManagedService) {
+			final Dictionary<String, Object> props = new Hashtable<>();
+			props.put(Constants.SERVICE_PID, PID);
+	        props.put(Constants.SERVICE_VENDOR, "The Apache Software Foundation");
+	        props.put(Constants.SERVICE_DESCRIPTION, "Managed Service for the Jetty Http Service");
+			this.configServiceReg = this.context.registerService("org.osgi.service.cm.ManagedService",
+	                new ServiceFactory()
+	                {
 
-        this.eventAdmintTracker = new ServiceTracker(this.context, EventAdmin.class.getName(), this);
-        this.eventAdmintTracker.open();
+	                    @Override
+	                    public Object getService(final Bundle bundle, final ServiceRegistration registration)
+	                    {
+	                        return new JettyManagedService(JettyService.this);
+	                    }
 
-        this.bundleTracker = new BundleTracker(this.context, Bundle.ACTIVE | Bundle.STARTING, this);
-        this.bundleTracker.open();
+	                    @Override
+	                    public void ungetService(Bundle bundle, ServiceRegistration registration, Object service)
+	                    {
+	                        // nothing to do
+	                    }
+	                }, props);
+        }
+
+        try {
+            this.webAppTracker = new WebAppBundleTracker(context, config);
+        } catch (Throwable t) {
+            SystemLogger.error("WebApp Bundle support not available: " + t.getMessage(), null);
+        }
+        if (this.webAppTracker != null) {
+            this.webAppTracker.start(parent);
+        }
     }
 
     public void stop() throws Exception
@@ -154,40 +154,28 @@ public final class JettyService extends AbstractLifeCycle.AbstractLifeCycleListe
             this.configServiceReg.unregister();
             this.configServiceReg = null;
         }
-        if (this.bundleTracker != null)
+        if (this.webAppTracker != null)
         {
-            this.bundleTracker.close();
-            this.bundleTracker = null;
-        }
-        if (this.eventAdmintTracker != null)
-        {
-            this.eventAdmintTracker.close();
-            this.eventAdmintTracker = null;
+            this.webAppTracker.stop();
+            this.webAppTracker = null;
         }
 
         // FELIX-4422: stop Jetty synchronously...
         stopJetty();
-
-        if (isExecutorServiceAvailable())
-        {
-            this.executor.shutdown();
-            // FELIX-4423: make sure to await the termination of the executor...
-            this.executor.awaitTermination(5, TimeUnit.SECONDS);
-        }
     }
 
-    private void publishServiceProperties()
+    private Hashtable<String, Object> getServiceProperties()
     {
-        Hashtable<String, Object> props = new Hashtable<String, Object>();
+        Hashtable<String, Object> props = new Hashtable<>();
         // Add some important configuration properties...
         this.config.setServiceProperties(props);
         addEndpointProperties(props, null);
 
         // propagate the new service properties to the actual HTTP service...
-        this.controller.setProperties(props);
+        return props;
     }
 
-    public void updated(Dictionary props)
+    public void updated(final Dictionary<String, ?> props)
     {
         if (this.config.update(props))
         {
@@ -213,19 +201,37 @@ public final class JettyService extends AbstractLifeCycle.AbstractLifeCycleListe
     {
         if (this.server != null)
         {
-            this.eventDispatcher.setActive(false);
+            this.controller.getEventDispatcher().setActive(false);
+            this.controller.unregister();
+
+            if (this.fileRequestLog != null)
+            {
+                this.fileRequestLog.stop();
+                this.fileRequestLog = null;
+            }
+            if (this.osgiRequestLog != null)
+            {
+                this.osgiRequestLog.unregister();
+                this.osgiRequestLog = null;
+            }
+            if (this.requestLogTracker != null)
+            {
+                this.requestLogTracker.close();
+                this.requestLogTracker = null;
+            }
+
             if (this.connectorTracker != null)
             {
                 this.connectorTracker.close();
                 this.connectorTracker = null;
             }
-            
+
             if (this.loadBalancerCustomizerTracker != null)
             {
                 this.loadBalancerCustomizerTracker.close();
                 this.loadBalancerCustomizerTracker = null;
-            } 
-            
+            }
+
             try
             {
                 this.server.stop();
@@ -249,7 +255,6 @@ public final class JettyService extends AbstractLifeCycle.AbstractLifeCycleListe
     {
         if (this.config.isUseHttp() || this.config.isUseHttps())
         {
-            final String version = fixJettyVersion();
 
             final int threadPoolMax = this.config.getThreadPoolMax();
             if (threadPoolMax >= 0) {
@@ -259,17 +264,24 @@ public final class JettyService extends AbstractLifeCycle.AbstractLifeCycleListe
             }
             this.server.addLifeCycleListener(this);
 
-            this.server.addBean(new HashLoginService("OSGi HTTP Service Realm"));
+            // FELIX-5931 : PropertyUserStore used as default by HashLoginService has changed in 9.4.12.v20180830
+            //              and fails without a config, therefore using plain UserStore
+            final HashLoginService loginService = new HashLoginService("OSGi HTTP Service Realm");
+            loginService.setUserStore(new UserStore());
+            this.server.addBean(loginService);
 
             this.parent = new ContextHandlerCollection();
 
-            ServletContextHandler context = new ServletContextHandler(this.parent, this.config.getContextPath(), ServletContextHandler.SESSIONS);
+            ServletContextHandler context = new ServletContextHandler(this.parent,
+                    this.config.getContextPath(),
+                    ServletContextHandler.SESSIONS);
 
             configureSessionManager(context);
-            this.eventDispatcher.setActive(true);
-            context.addEventListener(eventDispatcher);
-            context.getSessionHandler().addEventListener(eventDispatcher);
-            final ServletHolder holder = new ServletHolder(this.dispatcher);
+            this.controller.getEventDispatcher().setActive(true);
+            context.addEventListener(controller.getEventDispatcher());
+            context.getSessionHandler().addEventListener(controller.getEventDispatcher());
+
+            final ServletHolder holder = new ServletHolder(this.controller.createDispatcherServlet());
             holder.setAsyncSupported(true);
             context.addServlet(holder, "/*");
             context.setMaxFormContentSize(this.config.getMaxFormSize());
@@ -278,20 +290,58 @@ public final class JettyService extends AbstractLifeCycle.AbstractLifeCycleListe
             {
                 this.mbeanServerTracker = new MBeanServerTracker(this.context, this.server);
                 this.mbeanServerTracker.open();
-                context.addBean(new StatisticsHandler());
+                if (!this.config.isStatisticsHandlerEnabled()) {
+                  context.addBean(new StatisticsHandler());
+                }
             }
 
             this.server.setHandler(this.parent);
+
+            if (this.config.isStatisticsHandlerEnabled()) {
+              StatisticsHandler statisticsHandler = new StatisticsHandler();
+              this.server.insertHandler(statisticsHandler);
+              if (this.config.isRegisterMBeans())
+              {
+                context.addBean(statisticsHandler);
+              }
+            }
+
+            if (this.config.isGzipHandlerEnabled())
+            {
+            	GzipHandler gzipHandler = new GzipHandler();
+            	gzipHandler.setMinGzipSize(this.config.getGzipMinGzipSize());
+            	gzipHandler.setInflateBufferSize(this.config.getGzipInflateBufferSize());
+            	gzipHandler.setSyncFlush(this.config.isGzipSyncFlush());
+            	gzipHandler.addIncludedMethods(this.config.getGzipIncludedMethods());
+            	gzipHandler.addExcludedMethods(this.config.getGzipExcludedMethods());
+            	gzipHandler.addIncludedPaths(this.config.getGzipIncludedPaths());
+            	gzipHandler.addExcludedPaths(this.config.getGzipExcludedPaths());
+            	gzipHandler.addIncludedMimeTypes(this.config.getGzipIncludedMimeTypes());
+            	gzipHandler.addExcludedMimeTypes(this.config.getGzipExcludedMimeTypes());
+
+            	this.server.insertHandler(gzipHandler);
+            }
+
+            if(this.config.getStopTimeout() != -1)
+            {
+                this.server.setStopTimeout(this.config.getStopTimeout());
+            }
+
             this.server.start();
 
-            if (this.config.isProxyLoadBalancerConnection()) 
+            // session id manager is only available after server is started
+            context.getSessionHandler().getSessionIdManager().getSessionHouseKeeper().setIntervalSec(
+                    this.config.getLongProperty(JettyConfig.FELIX_JETTY_SESSION_SCAVENGING_INTERVAL,
+                            HouseKeeper.DEFAULT_PERIOD_MS / 1000L));
+
+            if (this.config.isProxyLoadBalancerConnection())
             {
                 customizerWrapper = new CustomizerWrapper();
                 this.loadBalancerCustomizerTracker = new LoadBalancerCustomizerFactoryTracker(this.context, customizerWrapper);
                 this.loadBalancerCustomizerTracker.open();
             }
-            
-            final StringBuilder message = new StringBuilder("Started Jetty ").append(version).append(" at port(s)");
+
+            final StringBuilder message = new StringBuilder("Started Jetty ").append(this.jettyVersion).append(" at port(s)");
             if (this.config.isUseHttp() && initializeHttp())
             {
                 message.append(" HTTP:").append(this.config.getHttpPort());
@@ -304,7 +354,7 @@ public final class JettyService extends AbstractLifeCycle.AbstractLifeCycleListe
 
             this.connectorTracker = new ConnectorFactoryTracker(this.context, this.server);
             this.connectorTracker.open();
-            
+
             if (this.server.getConnectors() != null && this.server.getConnectors().length > 0)
             {
                 message.append(" on context path ").append(this.config.getContextPath());
@@ -318,6 +368,7 @@ public final class JettyService extends AbstractLifeCycle.AbstractLifeCycleListe
                 }
                 Connector connector = this.server.getConnectors()[0];
                 if (connector instanceof ServerConnector) {
+                    @SuppressWarnings("resource")
                     ServerConnector serverConnector = (ServerConnector) connector;
                     message.append("acceptors=").append(serverConnector.getAcceptors()).append(",");
                     message.append("selectors=").append(serverConnector.getSelectorManager().getSelectorCount());
@@ -325,13 +376,33 @@ public final class JettyService extends AbstractLifeCycle.AbstractLifeCycleListe
                 message.append("]");
 
                 SystemLogger.info(message.toString());
-                publishServiceProperties();
+                this.controller.register(context.getServletContext(), getServiceProperties());
             }
             else
             {
                 this.stopJetty();
                 SystemLogger.error("Jetty stopped (no connectors available)", null);
             }
+
+            try {
+                this.requestLogTracker = new RequestLogTracker(this.context, this.config.getRequestLogFilter());
+                this.requestLogTracker.open();
+                this.server.setRequestLog(requestLogTracker);
+            } catch (InvalidSyntaxException e) {
+                SystemLogger.error("Invalid filter syntax in request log tracker", e);
+            }
+
+            if (this.config.isRequestLogOSGiEnabled()) {
+                this.osgiRequestLog = new LogServiceRequestLog(this.config);
+                this.osgiRequestLog.register(this.context);
+                SystemLogger.info("Directing Jetty request logs to the OSGi Log Service");
+            }
+
+            if (this.config.getRequestLogFilePath() != null && !this.config.getRequestLogFilePath().isEmpty()) {
+                this.fileRequestLog = new FileRequestLog(config);
+                this.fileRequestLog.start(this.context);
+                SystemLogger.info("Directing Jetty request logs to " + this.config.getRequestLogFilePath());
+            }
         }
         else
         {
@@ -339,11 +410,11 @@ public final class JettyService extends AbstractLifeCycle.AbstractLifeCycleListe
         }
     }
 
-    private String fixJettyVersion()
+    private static String fixJettyVersion(final BundleContext ctx)
     {
         // FELIX-4311: report the real version of Jetty...
-        Dictionary headers = this.context.getBundle().getHeaders();
-        String version = (String) headers.get("X-Jetty-Version");
+        final Dictionary<String, String> headers = ctx.getBundle().getHeaders();
+        String version = headers.get("X-Jetty-Version");
         if (version != null)
         {
             System.setProperty("jetty.version", version);
@@ -369,7 +440,7 @@ public final class JettyService extends AbstractLifeCycle.AbstractLifeCycleListe
 
         configureConnector(connector, this.config.getHttpPort());
 
-        if (this.config.isProxyLoadBalancerConnection()) 
+        if (this.config.isProxyLoadBalancerConnection())
         {
             connFactory.getHttpConfiguration().addCustomizer(customizerWrapper);
         }
@@ -381,7 +452,7 @@ public final class JettyService extends AbstractLifeCycle.AbstractLifeCycleListe
         HttpConnectionFactory connFactory = new HttpConnectionFactory();
         configureHttpConnectionFactory(connFactory);
 
-        SslContextFactory sslContextFactory = new SslContextFactory();
+        SslContextFactory.Server sslContextFactory = new SslContextFactory.Server();
         configureSslContextFactory(sslContextFactory);
 
         ServerConnector connector = new ServerConnector(
@@ -395,16 +466,37 @@ public final class JettyService extends AbstractLifeCycle.AbstractLifeCycleListe
         HttpConfiguration httpConfiguration = connFactory.getHttpConfiguration();
         httpConfiguration.addCustomizer(new SecureRequestCustomizer());
 
-        if (this.config.isProxyLoadBalancerConnection()) 
+        if (this.config.isProxyLoadBalancerConnection())
         {
             httpConfiguration.addCustomizer(customizerWrapper);
         }
-        
+
+        if (this.config.isUseHttp2()) {
+            //add ALPN factory
+            SslConnectionFactory alpnConnFactory = new SslConnectionFactory(sslContextFactory, "alpn");
+            connector.addConnectionFactory(alpnConnFactory);
+
+            ALPNServerConnectionFactory alpn = new ALPNServerConnectionFactory(this.config.getAlpnProtocols());
+            alpn.setDefaultProtocol(this.config.getAlpnDefaultProtocol());
+            connector.addConnectionFactory(alpn);
+
+            //Configure a HTTP2 on the ssl connector
+            HTTP2ServerConnectionFactory http2factory = new HTTP2ServerConnectionFactory(httpConfiguration);
+            http2factory.setMaxConcurrentStreams(this.config.getHttp2MaxConcurrentStreams());
+            http2factory.setInitialStreamRecvWindow(this.config.getHttp2InitialStreamRecvWindow());
+            http2factory.setInitialSessionRecvWindow(this.config.getHttp2InitialSessionRecvWindow());
+            connector.addConnectionFactory(http2factory);
+
+            //use http/2 cipher comparator
+            sslContextFactory.setCipherComparator(org.eclipse.jetty.http2.HTTP2Cipher.COMPARATOR);
+            sslContextFactory.setUseCipherSuitesOrder(true);
+        }
+
         configureConnector(connector, this.config.getHttpsPort());
         return startConnector(connector);
     }
 
-    private void configureSslContextFactory(final SslContextFactory connector)
+    private void configureSslContextFactory(final SslContextFactory.Server connector)
     {
         if (this.config.getKeystoreType() != null)
         {
@@ -481,7 +573,7 @@ public final class JettyService extends AbstractLifeCycle.AbstractLifeCycleListe
 
         if (this.config.isRegisterMBeans())
         {
-            connector.addBean(new ConnectorStatistics());
+            connector.addBean(new ConnectionStatistics());
         }
     }
 
@@ -498,25 +590,19 @@ public final class JettyService extends AbstractLifeCycle.AbstractLifeCycleListe
         config.setSendXPoweredBy(this.config.isSendServerHeader());
 
         connFactory.setInputBufferSize(this.config.getRequestBufferSize());
-
-        //Changed from 8.x to 9.x
-        //maxIdleTime -> ServerConnector.setIdleTimeout
-        //requestBufferSize -> HttpConnectionFactory.setInputBufferSize
-        //statsOn -> ServerConnector.addBean(new ConnectorStatistics());
     }
 
-    private void configureSessionManager(final ServletContextHandler context)
+    private void configureSessionManager(final ServletContextHandler context) throws Exception
     {
-        final SessionManager manager = context.getSessionHandler().getSessionManager();
+        final SessionHandler sessionHandler = context.getSessionHandler();
+        sessionHandler.setMaxInactiveInterval(this.config.getSessionTimeout() * 60);
+        sessionHandler.setSessionIdPathParameterName(this.config.getProperty(JettyConfig.FELIX_JETTY_SERVLET_SESSION_ID_PATH_PARAMETER_NAME, SessionHandler.__DefaultSessionIdPathParameterName));
+        sessionHandler.setCheckingRemoteSessionIdEncoding(this.config.getBooleanProperty(JettyConfig.FELIX_JETTY_SERVLET_CHECK_REMOTE_SESSION_ENCODING, true));
+        sessionHandler.setSessionTrackingModes(Collections.singleton(SessionTrackingMode.COOKIE));
 
-        manager.setMaxInactiveInterval(this.config.getSessionTimeout() * 60);
-        manager.setSessionIdPathParameterName(this.config.getProperty(JettyConfig.FELIX_JETTY_SERVLET_SESSION_ID_PATH_PARAMETER_NAME, SessionManager.__DefaultSessionIdPathParameterName));
-        manager.setCheckingRemoteSessionIdEncoding(this.config.getBooleanProperty(JettyConfig.FELIX_JETTY_SERVLET_CHECK_REMOTE_SESSION_ENCODING, true));
-        manager.setSessionTrackingModes(Collections.singleton(SessionTrackingMode.COOKIE));
-
-        SessionCookieConfig cookieConfig = manager.getSessionCookieConfig();
-        cookieConfig.setName(this.config.getProperty(JettyConfig.FELIX_JETTY_SERVLET_SESSION_COOKIE_NAME, SessionManager.__DefaultSessionCookie));
-        cookieConfig.setDomain(this.config.getProperty(JettyConfig.FELIX_JETTY_SERVLET_SESSION_DOMAIN, SessionManager.__DefaultSessionDomain));
+        final SessionCookieConfig cookieConfig = sessionHandler.getSessionCookieConfig();
+        cookieConfig.setName(this.config.getProperty(JettyConfig.FELIX_JETTY_SERVLET_SESSION_COOKIE_NAME, SessionHandler.__DefaultSessionCookie));
+        cookieConfig.setDomain(this.config.getProperty(JettyConfig.FELIX_JETTY_SERVLET_SESSION_DOMAIN, SessionHandler.__DefaultSessionDomain));
         cookieConfig.setPath(this.config.getProperty(JettyConfig.FELIX_JETTY_SERVLET_SESSION_PATH, context.getContextPath()));
         cookieConfig.setMaxAge(this.config.getIntProperty(JettyConfig.FELIX_JETTY_SERVLET_SESSION_MAX_AGE, -1));
         cookieConfig.setHttpOnly(this.config.getBooleanProperty(JettyConfig.FELIX_JETTY_SESSION_COOKIE_HTTP_ONLY, true));
@@ -599,7 +685,7 @@ public final class JettyService extends AbstractLifeCycle.AbstractLifeCycleListe
 
     private List<String> getEndpoints(final Connector connector, final List<NetworkInterface> interfaces)
     {
-        final List<String> endpoints = new ArrayList<String>();
+        final List<String> endpoints = new ArrayList<>();
         for (final NetworkInterface ni : interfaces)
         {
             final Enumeration<InetAddress> ias = ni.getInetAddresses();
@@ -618,7 +704,7 @@ public final class JettyService extends AbstractLifeCycle.AbstractLifeCycleListe
 
     private void addEndpointProperties(final Hashtable<String, Object> props, Object container)
     {
-        final List<String> endpoints = new ArrayList<String>();
+        final List<String> endpoints = new ArrayList<>();
 
         final Connector[] connectors = this.server.getConnectors();
         if (connectors != null)
@@ -627,12 +713,12 @@ public final class JettyService extends AbstractLifeCycle.AbstractLifeCycleListe
             {
                 final Connector connector = connectors[i];
 
-                if (getServerConnector(connector).getHost() == null)
+                if (getServerConnector(connector).getHost() == null || "0.0.0.0".equals(getServerConnector(connector).getHost()))
                 {
                     try
                     {
-                        final List<NetworkInterface> interfaces = new ArrayList<NetworkInterface>();
-                        final List<NetworkInterface> loopBackInterfaces = new ArrayList<NetworkInterface>();
+                        final List<NetworkInterface> interfaces = new ArrayList<>();
+                        final List<NetworkInterface> loopBackInterfaces = new ArrayList<>();
                         final Enumeration<NetworkInterface> nis = NetworkInterface.getNetworkInterfaces();
                         while (nis.hasMoreElements())
                         {
@@ -676,308 +762,19 @@ public final class JettyService extends AbstractLifeCycle.AbstractLifeCycleListe
                 endpoints.toArray(new String[endpoints.size()]));
     }
 
-    private Deployment startWebAppBundle(Bundle bundle, String contextPath)
-    {
-        postEvent(WebEvent.DEPLOYING(bundle, this.context.getBundle()));
-
-        // check existing deployments
-        Deployment deployment = this.deployments.get(contextPath);
-        if (deployment != null)
-        {
-            SystemLogger.warning(String.format("Web application bundle %s has context path %s which is already registered", bundle.getSymbolicName(), contextPath), null);
-            postEvent(WebEvent.FAILED(bundle, this.context.getBundle(), null, contextPath, deployment.getBundle().getBundleId()));
-            return null;
-        }
-
-        // check context path belonging to Http Service implementation
-        if (contextPath.equals("/"))
-        {
-            SystemLogger.warning(String.format("Web application bundle %s has context path %s which is reserved", bundle.getSymbolicName(), contextPath), null);
-            postEvent(WebEvent.FAILED(bundle, this.context.getBundle(), null, contextPath, this.context.getBundle().getBundleId()));
-            return null;
-        }
-
-        // check against excluded paths
-        for (String path : this.config.getPathExclusions())
-        {
-            if (contextPath.startsWith(path))
-            {
-                SystemLogger.warning(String.format("Web application bundle %s has context path %s which clashes with excluded path prefix %s", bundle.getSymbolicName(), contextPath, path), null);
-                postEvent(WebEvent.FAILED(bundle, this.context.getBundle(), null, path, null));
-                return null;
-            }
-        }
-
-        deployment = new Deployment(contextPath, bundle);
-        this.deployments.put(contextPath, deployment);
-
-        WebAppBundleContext context = new WebAppBundleContext(contextPath, bundle, this.getClass().getClassLoader());
-        deploy(deployment, context);
-        return deployment;
-    }
-
-    public void deploy(final Deployment deployment, final WebAppBundleContext context)
-    {
-        if (!isExecutorServiceAvailable())
-        {
-            // Shutting down...?
-            return;
-        }
-
-        this.executor.submit(new JettyOperation()
-        {
-            @Override
-            protected void doExecute()
-            {
-                final Bundle webAppBundle = deployment.getBundle();
-                final Bundle extenderBundle = JettyService.this.context.getBundle();
-
-                try
-                {
-                    context.getServletContext().setAttribute(OSGI_BUNDLE_CONTEXT, webAppBundle.getBundleContext());
-
-                    JettyService.this.parent.addHandler(context);
-                    context.start();
-
-                    Dictionary<String, Object> props = new Hashtable<String, Object>();
-                    props.put(WEB_SYMBOLIC_NAME, webAppBundle.getSymbolicName());
-                    props.put(WEB_VERSION, webAppBundle.getVersion());
-                    props.put(WEB_CONTEXT_PATH, deployment.getContextPath());
-                    deployment.setRegistration(webAppBundle.getBundleContext().registerService(ServletContext.class.getName(), context.getServletContext(), props));
-
-                    postEvent(WebEvent.DEPLOYED(webAppBundle, extenderBundle));
-                }
-                catch (Exception e)
-                {
-                    SystemLogger.error(String.format("Deploying web application bundle %s failed.", webAppBundle.getSymbolicName()), e);
-                    postEvent(WebEvent.FAILED(webAppBundle, extenderBundle, e, null, null));
-                    deployment.setContext(null);
-                }
-            }
-        });
-        deployment.setContext(context);
-    }
-
-    public void undeploy(final Deployment deployment, final WebAppBundleContext context)
-    {
-        if (!isExecutorServiceAvailable())
-        {
-            // Already stopped...?
-            return;
-        }
-
-        this.executor.submit(new JettyOperation()
-        {
-            @Override
-            protected void doExecute()
-            {
-                final Bundle webAppBundle = deployment.getBundle();
-                final Bundle extenderBundle = JettyService.this.context.getBundle();
-
-                try
-                {
-                    postEvent(WebEvent.UNDEPLOYING(webAppBundle, extenderBundle));
-
-                    context.getServletContext().removeAttribute(OSGI_BUNDLE_CONTEXT);
-
-                    ServiceRegistration registration = deployment.getRegistration();
-                    if (registration != null)
-                    {
-                        registration.unregister();
-                    }
-                    deployment.setRegistration(null);
-                    deployment.setContext(null);
-                    context.stop();
-                }
-                catch (Exception e)
-                {
-                    SystemLogger.error(String.format("Undeploying web application bundle %s failed.", webAppBundle.getSymbolicName()), e);
-                }
-                finally
-                {
-                    postEvent(WebEvent.UNDEPLOYED(webAppBundle, extenderBundle));
-                }
-            }
-        });
-    }
-
     @Override
-    public Object addingBundle(Bundle bundle, BundleEvent event)
+    public void lifeCycleStarted(final LifeCycle event)
     {
-        return detectWebAppBundle(bundle);
-    }
-
-    @Override
-    public void modifiedBundle(Bundle bundle, BundleEvent event, Object object)
-    {
-        detectWebAppBundle(bundle);
-    }
-
-    private Object detectWebAppBundle(Bundle bundle)
-    {
-        if (bundle.getState() == Bundle.ACTIVE || (bundle.getState() == Bundle.STARTING && "Lazy".equals(bundle.getHeaders().get(HEADER_ACTIVATION_POLICY))))
-        {
-
-            String contextPath = bundle.getHeaders().get(HEADER_WEB_CONTEXT_PATH);
-            if (contextPath != null)
-            {
-                return startWebAppBundle(bundle, contextPath);
-            }
-        }
-        return null;
-    }
-
-    @Override
-    public void removedBundle(Bundle bundle, BundleEvent event, Object object)
-    {
-        String contextPath = bundle.getHeaders().get(HEADER_WEB_CONTEXT_PATH);
-        if (contextPath == null)
-        {
-            return;
-        }
-
-        Deployment deployment = this.deployments.remove(contextPath);
-        if (deployment != null && deployment.getContext() != null)
-        {
-            // remove registration, since bundle is already stopping
-            deployment.setRegistration(null);
-            undeploy(deployment, deployment.getContext());
+        if (this.webAppTracker != null) {
+            this.webAppTracker.lifeCycleStarted(event);
         }
     }
 
     @Override
-    public Object addingService(ServiceReference reference)
+    public void lifeCycleStopping(final LifeCycle event)
     {
-        Object service = this.context.getService(reference);
-        modifiedService(reference, service);
-        return service;
-    }
-
-    @Override
-    public void modifiedService(ServiceReference reference, Object service)
-    {
-        this.eventAdmin = (EventAdmin) service;
-    }
-
-    @Override
-    public void removedService(ServiceReference reference, Object service)
-    {
-        this.context.ungetService(reference);
-        this.eventAdmin = null;
-    }
-
-    private void postEvent(Event event)
-    {
-        if (this.eventAdmin != null)
-        {
-            this.eventAdmin.postEvent(event);
+        if (this.webAppTracker != null) {
+            this.webAppTracker.lifeCycleStopping(event);
         }
     }
-
-    @Override
-    public void lifeCycleStarted(LifeCycle event)
-    {
-        for (Deployment deployment : this.deployments.values())
-        {
-            if (deployment.getContext() == null)
-            {
-                postEvent(WebEvent.DEPLOYING(deployment.getBundle(), this.context.getBundle()));
-                WebAppBundleContext context = new WebAppBundleContext(deployment.getContextPath(), deployment.getBundle(), this.getClass().getClassLoader());
-                deploy(deployment, context);
-            }
-        }
-    }
-
-    @Override
-    public void lifeCycleStopping(LifeCycle event)
-    {
-        for (Deployment deployment : this.deployments.values())
-        {
-            if (deployment.getContext() != null)
-            {
-                undeploy(deployment, deployment.getContext());
-            }
-        }
-    }
-
-    /**
-     * A deployment represents a web application bundle that may or may not be deployed.
-     */
-    static class Deployment
-    {
-        private String contextPath;
-        private Bundle bundle;
-        private WebAppBundleContext context;
-        private ServiceRegistration registration;
-
-        public Deployment(String contextPath, Bundle bundle)
-        {
-            this.contextPath = contextPath;
-            this.bundle = bundle;
-        }
-
-        public Bundle getBundle()
-        {
-            return this.bundle;
-        }
-
-        public String getContextPath()
-        {
-            return this.contextPath;
-        }
-
-        public WebAppBundleContext getContext()
-        {
-            return this.context;
-        }
-
-        public void setContext(WebAppBundleContext context)
-        {
-            this.context = context;
-        }
-
-        public ServiceRegistration getRegistration()
-        {
-            return this.registration;
-        }
-
-        public void setRegistration(ServiceRegistration registration)
-        {
-            this.registration = registration;
-        }
-    }
-
-    /**
-     * A Jetty operation is executed with the context class loader set to this class's
-     * class loader.
-     */
-    abstract static class JettyOperation implements Callable<Void>
-    {
-        @Override
-        public Void call() throws Exception
-        {
-            ClassLoader cl = Thread.currentThread().getContextClassLoader();
-
-            try
-            {
-                Thread.currentThread().setContextClassLoader(this.getClass().getClassLoader());
-                doExecute();
-                return null;
-            }
-            finally
-            {
-                Thread.currentThread().setContextClassLoader(cl);
-            }
-        }
-
-        protected abstract void doExecute() throws Exception;
-    }
-
-    /**
-     * @return <code>true</code> if there is a valid executor service available, <code>false</code> otherwise.
-     */
-    private boolean isExecutorServiceAvailable()
-    {
-        return this.executor != null && !this.executor.isShutdown();
-    }
 }
